unit module_defs;

{

  (C) COPYRIGHT 1999 .. 2014 Blue Hell / Jan Punter

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License version 2 as
  published by the Free Software Foundation;

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  For all listed email addresses :

    _dot. to be substituted by a dot      '.'
    2@t2  to be substituted by an at sign '@'


  Blue Hell is a trade mark owned by

    Jan Punter
    Oogstplein 6
    7545 HP Enschede
    the Netherlands
    http://www.iaf.nl/Users/BlueHell/
    http://bluehell.electro-music.com/
    j_dot.punter2@t2iaf_dot.nl

  All rights attributed to Blue Hell are owned by Jan Punter.
}

interface

uses

  System.SysUtils, System.Math, System.Classes;

const

  PI_DIV_2          = PI / 2;                        // Offset for cosine lookups - radian based
  PI_2              = PI * 2;                        // Two times Pi, for angle conversions
  DELTA             = 1e-6;                          // Small value for equality comparisons

  SYSTEM_RATE       = 44100.0;                       // samples / s
  SYSTEM_RATE_2     = SYSTEM_RATE * 2;               // samples / s
  SLOW_DECIMATION   = 4;                             // Decimation used for slow tick stuff (control rate)
  SLOW_RATE         = SYSTEM_RATE / SLOW_DECIMATION;

  MAX_SIGNAL        = 1e3;                           // Maximum signal level in a patch - for anti runaway
  MIN_SIGNAL        = - MAX_SIGNAL;                  // Minumum signal level in a patch - for anti runaway

  NO_INPUT          = 'noinput';
  NO_OUTPUT         = 'nooutput';
  NO_INTERNAL       = 'nointernal';
  NO_STRING         = 'nostring';
  NO_MODULE         = 'nomodule';
  PARENT_INDEX      = -2;                            // Module index -2 denotes the container for a module
  UNDEF_POMOC       = -1;                            // Pin/module/connection index for non existent pin/module/connection


  {$IFDEF DEBUG}
    UNDEFINED = NAN;                               // NAN for stability checks
    {$INLINE OFF}
  {$ELSE}
    UNDEFINED = 0.0;                               // 0.0 for production code
    {$INLINE ON}
  {$ENDIF}

  WhiteSpace         = [ #$00 .. #$20, #$7f];
  StrictDigits       = [ '0' .. '9'];
  Digits             = StrictDigits + [ '+', '-'];
  IdentifierStarters = [ 'A' .. 'Z', 'a' .. 'z', '%', '#', '@', '_'];
  IdentifierChars    = IdentifierStarters + Digits + [ '&'];
  FloatChars         = Digits + [ '.', 'e', 'E'];


  // Midi command definitions

  CNoteOff         = $80; // Length = 3;
  CNoteOn          = $90; // Length = 3;
  CKeyPressure     = $a0; // Length = 3;
  CControlChange   = $b0; // Length = 3;
  CProgramChange   = $c0; // Length = 2;
  CChannelPressure = $d0; // Length = 2;
  CPitchBend       = $e0; // Length = 2;
  CSysEx           = $f0; // Length = n;

  // When aCmd And $f0 = CSysEx low nibble of 1st byte is :

  SSysEx           = $00; // Length = 1, followed by arbitrary amt of data, finalized by SEox;
  SSongPosition    = $02; // Length = 3;
  SSongSelect      = $03; // Length = 2;
  STuneRequest     = $06; // Length = 1;
  SEox             = $07; // Length = 1;
  STimingClock     = $08; // Length = 1;
  SStart           = $0a; // Length = 1;
  SContinue        = $0b; // Length = 1;
  SStop            = $0c; // Length = 1;
  SActiveSense     = $0e; // Length = 1;
  SReset           = $0f; // Length = 1;

  CCAllNotesOff    = 123; // Pseudo CC - all notes off

type

  EModule            = class( Exception);
  EModuleNotFound    = class( EModule  );
  EInputNotFound     = class( EModule  );
  EOutputNotFound    = class( EModule  );
  EInternalNotFound  = class( EModule  );
  EStringNotFound    = class( EModule  );
  EUnitsUndefined    = class( EModule  );

  TStringArray       = array of string;
  TSignal            = Double;
  TSignalArray       = array of TSignal;
  TConnection        = class;
  TConnectionArray   = array of TConnection;
  TMod               = class;
  TSynthPatch        = class;
  TModuleArray       = array of TMod;
  TModuleClass       = class of TMod;

  TInts              = array of Integer;

  TMidiMessage = record
    Command : Byte;
    Channel : Byte;
    Data    : TInts;
  end;

  TLightsHandler     = procedure( aSender: TMod; const aMsg: string)                             of object;
  TDataHandler       = procedure( aSender: TMod; const aPath: string; const aData: TSignalArray) of object;
  TInfoHandler       = procedure( aSender: TMod; const aMsg: string)                             of object;
  TOnSendMidi        = procedure( aSender: TObject; const aMsg: TMidiMessage)                    of object;

  TSignals = class
  private
    FSignals: TSignalArray;
  private
    function    GetCount        : Integer;
    function    GetSignal       ( anIndex: Integer): TSignal;
    procedure   SetSignal       ( anIndex: Integer; aValue: TSignal);
    function    GetInterpolate  ( anIndex: TSignal): TSignal;
  public
    constructor Create( aCount: Integer);                                           overload;
    destructor  Destroy;                                                            override;
    procedure   Clear;
    procedure   AddSignal( aValue: TSignal);
    procedure   SetSignals( aStartIndex: Integer; const aValue: TSignals);
    function    Convolve( aValue: TSignal): TSignal;
  public
    property    Count                         : Integer read GetCount;
    property    Signal     [ anIndex: Integer]: TSignal read GetSignal      write SetSignal;
    property    Interpolate[ anIndex: TSignal]: TSignal read GetInterpolate;
  end;

  TSignalPair = record
    Left  : TSignal;
    Right : TSignal;
  end;

  TSignalPairs = array of TSignalPair;

  TSignalQueue = class
  private
    FSignals    : TSignalPairs;
    FInp        : Integer;
    FOutp       : Integer;
    FFreeCount  : Integer;
    FFillCount  : Integer;
    FOverflows  : Integer;
    FUnderflows : Integer;
    FLastValue  : TSignalPair;
  private
    function    GetOverflows : Integer;
    function    GetUnderflows: Integer;
  public
    constructor Create( aSize: Integer);
    destructor  Destroy;                                 override;
    procedure   Clear;
    procedure   Append( aLeft, aRight: TSignal);         overload;
    procedure   Append( aPair: TSignalPair);             overload;
    function    GetData: TSignalPair;
  public
    property    Overflows : Integer read GetOverflows;
    property    Underflows: Integer read GetUnderflows;
  end;

  TConnection = class
  public
    Peer      : TSynthPatch;
    Index     : Integer;
    SrcModule : Integer;
    SrcPin    : Integer;
    DstModule : Integer;
    DstPin    : Integer;
  private
    procedure   Connect( const aPeer: TSynthPatch; anIndex, aSrcModule, aSrcPin, aDstModule, aDstPin: Integer);
  public
    procedure   Save( anIndent: Integer; var S: string);
    function    IsSameConnection( aConnection: TConnection): Boolean;
  end;

  TConnections = class
  private
    FPeer       : TSynthPatch;
    FConnections: TConnectionArray;
  private
    function    GetCount: Integer;
    function    GetConnection( anIndex: Integer): TConnection;
    procedure   SetConnection( anIndex: Integer; const aValue: TConnection);
  public
    constructor Create( aPeer: TSynthPatch);
    destructor  Destroy;                                                                     override;
    procedure   Save( anIndent: Integer; const aName: string; var S: string);
    procedure   Clear;
    procedure   AddConnection( const aValue: TConnection);                                   overload;
    procedure   AddConnection( aSrcModule, aSrcPin, aDstModule, aDstPin: Integer);           overload;
    procedure   RemoveConnection( anIndex: Integer);
    procedure   RemoveConnectionsFor( aModuleIndex: Integer);
    procedure   PropagateSignals( const aModules: TModuleArray);
    function    IsSameConnections( const aConnections: TConnections): Boolean;
  public
    property    Count: Integer read GetCount;
    property    Connection[ anIndex: Integer]: TConnection read GetConnection write SetConnection;
  end;

  TMod = class
  private
    FParent      : TSynthPatch;
    FName        : string;
    FInputs      : TSignals;
    FOutputs     : TSignals;
    FTickLocks   : Integer;
    FSamples     : Int64;
    FClocks      : Int64;
    FTotalClocks : Int64;
  protected
    FIsSlow   : Boolean;           // True when slow mode processing must be performed
    FIsFast   : Boolean;           // True when fast mode processing must be performed
    FIsMidi   : Boolean;           // True when midi reception must be enabled
    FInCache  : TSignalArray;
    FOutCache : TSignalArray;
  private
    function    AssertUniqueName( const aName: string): string;                                   virtual;
  private
    procedure   SetName( const aValue: string);
    function    GetInputCount: Integer;
    function    GetOutputCount: Integer;
    function    GetOutput( anIndex: Integer): TSignal;
    procedure   SetCommonValue        ( anIndex : Integer; aValue: TSignal);                      virtual;
    procedure   SetParamValue         ( const aName: string; aValue: TSignal);                    virtual;
    procedure   SetInternalValue      ( const aName: string; aValue: TSignal);                    virtual;
    procedure   SetInternalStringValue( const aName, aValue: string);
  protected
    procedure   SetStringValue        ( const aName, aValue: string);                             virtual;
    function    GetStringValue        ( const aName: string): string;                             virtual;
    procedure   FixOutputs;                                                                       virtual;
    procedure   SetOutputCount( aCount: Integer);
    procedure   SetInputs( aStartIndex: Integer; const aValue: TSignals);
    procedure   HandleParam( const aControl: string; aValue: TSignal; IsFinal: Boolean);          virtual;
    procedure   ReleaseTickLock;
    function    IsTickLocked: Boolean;
    function    GrabTickLock: Boolean;
  public
    constructor Create( aParent: TSynthPatch; const aName: string; anInputCount: Integer);        virtual;
    destructor  Destroy;                                                                         override;
    procedure   SetDefaults;                                                                      virtual;
    function    SensibleSignal( aValue: TSignal): string;
    procedure   SaveInputValues ( anIndent: Integer; var S: string);
    procedure   SaveKeyValue    ( anIndent: Integer; const aName: string; aValue: TSignal; var S: string);
    procedure   SaveDetails     ( anIndent: Integer; var S: string);                              virtual;
    procedure   SaveInternals   ( anIndent: Integer; var S: string);                              virtual;
    procedure   SaveStrings     ( anIndent: Integer; var S: string);                              virtual;
    procedure   SetInternal     ( const aName: string; aValue: TSignal);                          virtual;
    function    GetInternal     ( const aName: string): TSignal;                                  virtual;
    procedure   Save            ( anIndent: Integer; var S: string);
    procedure   Reset;                                                                            virtual;
    procedure   Clear;                                                                            virtual;
    function    FindInput   ( const aPinName: string): Integer;
    function    FindOutput  ( const aPinName: string): Integer;
    function    InputName   ( anIndex: Integer): string;                                          virtual;
    function    OutputName  ( anIndex: Integer): string;                                          virtual;
    function    InternalName( anIndex: Integer): string;                                          virtual;
    function    StringName  ( anIndex: Integer): string;                                          virtual;
    procedure   PreTick;                                                                           inline;
    procedure   PostTick;                                                                          inline;
    procedure   Tick;
    Procedure   SlowTick;
    procedure   DoTick;                                                                           virtual;
    procedure   DoSlowTick;                                                                       virtual;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);            virtual;
    procedure   GatherData  ( const aPrefix: string; const aCallback: TDataHandler);              virtual;
    procedure   GatherInfo  ( const aPrefix: string; const aCallback: TInfoHandler);              virtual;
    procedure   AcceptParam( const aPath: string; aValue: TSignal; IsFinal: Boolean);
    function    GetProfileInfoHtml: string;
    function    IsSameModule( const aModule: TMod): Boolean;
    procedure   AcceptMidi( const aMsg: TMidiMessage);                                            virtual;
    procedure   SendMidi( const aMsg: TMidiMessage);                                              virtual;
  public
    property    Name                             : string   read FName                  write SetName;
    property    TotalClocks                      : Int64    read FTotalClocks;
    property    Samples                          : Int64    read FSamples;
    property    InputCount                       : Integer  read GetInputCount;
    property    OutputCount                      : Integer  read GetOutputCount;
    property    Output[ anIndex: Integer]        : TSignal  read GetOutput;
  public
    property    ParamValue         [ const aName: string] : TSignal                     write SetParamValue;
    property    InternalValue      [ const aName: string] : TSignal                     write SetInternalValue;
    property    InternalStringValue[ const aName: string] : string                      write SetInternalStringValue;
    property    StringValue        [ const aName: string] : string  read GetStringValue write SetStringValue;
  end;

  TModListener = class( TMod)
  public
    procedure   AcceptText( const aValue: string); virtual; abstract;
    procedure   AcceptData( const aValue: TBytes); virtual; abstract;
  end;

  TListenerArray = array of TModListener;

  TSynthPatch = class( TModListener)
  private
    FVersion          : Integer;
    FModules          : TModuleArray;
    FConnections      : TConnections;
    FInputMappings    : TConnections;
    FOutputMappings   : TConnections;
    FListeners        : TListenerArray;
    FDecimatorCount   : Integer;
    FInputQueue       : TSignalQueue;
    FWireClocks       : Int64;
    FTotalWireClocks  : Int64;
    FOnSendMidi       : TOnSendMidi;
  private
    function    AssertUniqueName( const aName: string): string;                                  override;
  private
    function    CheckModuleIndex( anIndex: Integer): Boolean;
    function    GetModuleCount: Integer;
    function    GetModule( anIndex: Integer): TMod;
    function    GetConnectionCount: Integer;
    function    GetOverflows : Integer;
    function    GetUnderflows: Integer;
  protected
    procedure   HandleParam( const aPath: string; aValue: TSignal; IsFinal: Boolean);            override;
  public
    constructor Create( aParent: TSynthPatch; const aName: string; anInputCount: Integer);       override;
    destructor  Destroy;                                                                         override;
    procedure   SetDefaults;                                                                     override;
    procedure   AcceptText( const aValue: string);                                               override;
    procedure   AcceptData( const aValue: TBytes);                                               override;
    procedure   SaveInternals     ( anIndent: Integer; var S: string);                           override;
    procedure   SetInternal( const aName: string; aValue: TSignal);                              override;
    function    GetInternal( const aName: string): TSignal;                                      override;
    procedure   SaveConnections   ( anIndent: Integer; var S: string);
    procedure   SaveInputMappings ( anIndent: Integer; var S: string);
    procedure   SaveOutputMappings( anIndent: Integer; var S: string);
    procedure   SaveModules       ( anIndent: Integer; var S: string);
    procedure   SaveDetails       ( anIndent: Integer; var S: string);                           override;
    procedure   Clear;                                                                           override;
    procedure   Reset;                                                                           override;
    procedure   RemoveConnectionsFor( aModuleIndex: Integer);
    procedure   SignalModuleRemoval( const aModule: TMod);
    procedure   SignalModuleAddition( const aModule: TMod);
    procedure   ClearModules;
    function    FindModule( const aModuleName: string): Integer;
    function    ModuleName( anIndex: Integer): string;
    function    FindInputPin ( const aModuleIndex: Integer; const aPinName: string): Integer;
    function    FindOutputPin( const aModuleIndex: Integer; const aPinName: string): Integer;
    function    InputPinName ( aModuleIndex, aPinIndex: Integer): string;
    function    OutputPinName( aModuleIndex, aPinIndex: Integer): string;
    procedure   AddModule( const aModule: TMod);
    procedure   AddConnection( aSrcModule, aSrcPin, aDstModule, aDstPin: Integer);               overload;
    procedure   AddConnection( const aSrcModule, aSrcPin, aDstModule, aDstPin: string);          overload;
    procedure   AddInputMapping ( aSrcPin, aDstModule, aDstPin: Integer);                        overload;
    procedure   AddOutputMapping( aSrcModule, aSrcPin, aDstPin: Integer);                        overload;
    procedure   AddInputMapping ( aSrcPin: Integer; const aDstModule, aDstPin: string);          overload;
    procedure   AddOutputMapping( const aSrcModule, aSrcPin: string; aDstPin: Integer);          overload;
    procedure   PreWires;
    procedure   PostWires;
    procedure   PropagateSignals;
    procedure   ApplyInputMappings;                                                               virtual;
    procedure   ApplyOutputMappings;                                                              virtual;
    function    InternalName( anIndex: Integer): string;                                         override;
    procedure   DoTick;                                                                          override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    procedure   GatherData  ( const aPrefix: string; const aCallback: TDataHandler);             override;
    procedure   GatherInfo  ( const aPrefix: string; const aCallback: TInfoHandler);             override;
    procedure   AcceptSamplePair( aLeft, aRight: TSignal);
    function    GetProfileHtml: TStringList;
    function    IsSamePatch       ( const aPatch: TSynthPatch): Boolean;
    function    SameModules       ( const aPatch: TSynthPatch): Boolean;
    function    SameConnections   ( const aPatch: TSynthPatch): Boolean;
    function    SameInputMappings ( const aPatch: TSynthPatch): Boolean;
    function    SameOutputMappings( const aPatch: TSynthPatch): Boolean;
    procedure   AcceptMidi( const aMsg: TMidiMessage);                                           override;
    procedure   SendMidi( const aMsg: TMidiMessage);                                             override;
  public
    property    Version                      : Integer     read FVersion       write FVersion;
    property    ModuleCount                  : Integer     read GetModuleCount;
    property    Module    [ anIndex: Integer]: TMod        read GetModule;
    property    ConnectionCount              : Integer     read GetConnectionCount;
    property    Overflows                    : Integer     read GetOverflows;
    property    Underflows                   : Integer     read GetUnderflows;
    property    OnSendMidi                   : TOnSendMidi read FOnSendMidi    write FOnSendMidi;
  end;


  procedure RegisterModuleType( aModuleType: TModuleClass);
  procedure RegisterModuleTypes( aModuleTypes: array of TModuleClass);

  function  Clip( const aValue: TSignal; const aLow, aHigh : TSignal ): TSignal;    inline;
  function  RangeMap( const aValue, anInLow, anInHigh, anOutLow, anOutHigh: TSignal): TSignal; inline;
  function  AlmostEqual              ( aValue1, aValue2, aDelta: TSignal): Boolean; inline;
  function  FModulo                  ( aValue, aDivisor: TSignal): TSignal;         inline;
  function  AntiRunAway              ( aValue  : TSignal): TSignal;                 inline;
  function  AlmostZero               ( aValue  : TSignal): Boolean;                 inline;
  function  LogicToSignal            ( aValue  : Boolean): TSignal;                 inline;
  function  SignalToLogic            ( aValue  : TSignal): Boolean;                 inline;
  function  SignalToInt              ( aValue  : TSignal): Integer;                 inline;
  function  SignalTodB               ( aValue  : TSignal): TSignal;                 inline;

  function  NoteNumberToUnits        ( aValue  : TSignal): TSignal;                 inline;
  function  SignalToNoteNumber       ( aValue  : TSignal): TSignal;                 inline;
  function  UnitsToFrequency         ( aValue  : TSignal): TSignal;                 inline;
  function  UnitsToFrequencyLin      ( aValue  : TSignal): TSignal;                 inline;
  function  FrequencyToUnits         ( aValue  : TSignal): TSignal;                 inline;
  function  NoteNrToFrequency        ( aValue  : TSignal): TSignal;                 inline;
  function  FrequencyToNoteNr        ( aValue  : TSignal): TSignal;                 inline;
  function  FrequencyToPhaseDelta    ( aSignal : TSignal): TSignal;                 inline;
  function  SampleCountToTime        ( aCount  : TSignal): TSignal;                 inline;
  function  TimeToSampleCount        ( aTime   : TSignal): TSignal;                 inline;

  function  SlowUnitsToFrequency     ( aValue  : TSignal): TSignal;                 inline;
  function  SlowNoteNrToFrequency    ( aValue  : TSignal): TSignal;                 inline;
  function  SlowFrequencyToPhaseDelta( aSignal : TSignal): TSignal;                 inline;
  function  SlowTimeToSampleCount ( aTime   : TSignal): TSignal;                    inline;
  function  SlowSampleCountToTime ( aCount  : TSignal): TSignal;                    inline;

  function  NoteNameToUnits      ( const aNoteName : string; aDefault: TSignal): TSignal;
  function  UnitsToNoteName      ( aValue: TSignal; Compact: Boolean = True): string;
  function  IsNoteName           ( const aNoteName: string): Boolean;

  function  CreateModule( aParent: TSynthPatch; const aTypeName, aName: string; anInputCount: Integer): TMod;
  procedure ParsePrefix( const aPath: string; var aPrefix, aRest: string);
  function  ParsePostfix( const aPath: string): string;
  function  ApplyUnitsToSignal( var aValue: TSignal; const aUnits: string): Boolean;
  function  ParseTextValue( const aString: string; var aSignal: TSignal): Boolean;

  function  ParseTextSequencerValue( const aValue: string): TSignals;

  function  ExpAttack( aTime, aDuration, aStartValue, anEndValue: TSignal): TSignal;
  function  LinAttack( aTime, aDuration, aStartValue, anEndValue: TSignal): TSignal;
  function  LogAttack( aTime, aDuration, aStartValue, anEndValue: TSignal): TSignal;
  function  ExpDecay ( aTime, aDuration, aStartValue, anEndValue: TSignal): TSignal;
  function  LinDecay ( aTime, aDuration, aStartValue, anEndValue: TSignal): TSignal;
  function  LogDecay ( aTime, aDuration, aStartValue, anEndValue: TSignal): TSignal;

  function  MakeMidiNoteOff        ( aVelocity, aNote, aChannel: Byte)    : TMidiMessage;
  function  MakeMidiNoteOn         ( aVelocity, aNote, aChannel: Byte)    : TMidiMessage;
  function  MakeMidiKeyPressure    ( aPressure, aNote, aChannel: Byte)    : TMidiMessage;
  function  MakeMidiControlChange  ( aValue, aController, aChannel: Byte) : TMidiMessage;
  function  MakeMidiProgramChange  ( aProgram, aChannel: Byte)            : TMidiMessage;
  function  MakeMidiChannelPressure( aValue, aChannel: Byte)              : TMidiMessage;
  function  MakeMidiPitchBend      ( aData: Word; aChannel: Byte)         : TMidiMessage;
  function  MakeMidiSongPosition   ( aData: Word)                         : TMidiMessage;
  function  MakeMidiSongSelect     ( aSong: Byte)                         : TMidiMessage;
  function  MakeMidiTuneRequest                                           : TMidiMessage;
  function  MakeMidiTimingClock                                           : TMidiMessage;
  function  MakeMidiStart                                                 : TMidiMessage;
  function  MakeMidiContinue                                              : TMidiMessage;
  function  MakeMidiStop                                                  : TMidiMessage;
  function  MakeMidiActiveSense                                           : TMidiMessage;
  function  MakeMidiReset                                                 : TMidiMessage;
  function  MakeMidiSysExData      ( const aData: TBytes)                 : TMidiMessage;

  procedure MidiMessageToBytes( const aMsg: TMidiMessage; var aData: TBytes);
  function  MidiMessageToLog( const aMsg: TMidiMessage): string;


var

  ReferenceA     : TSignal;  // The value of the reference A to be used - default 440 Hz
  NotesPerOctave : TSignal;  // The number of notes wanted in an octave - there only is notation for 12 tho ... A .. G
  MiddleNote     : TSignal;  // The middle note to be used - default 69


implementation



uses

  Globals;

type

  TModuleTypes = class( TStringList)
  public
    constructor Create;
    procedure   RegisterType( aModuleType: TModuleClass);
    function    CreateModule( aParent: TSynthPatch; const aTypeName, aName: string; anInputCount: Integer): TMod;
  end;


var

  FModuleTypes: TModuleTypes;


  function CreateModule( aParent: TSynthPatch; const aTypeName, aName: string; anInputCount: Integer): TMod;
  begin
    Result := FModuleTypes.CreateModule( aParent, aTypeName, aName, anInputCount);
  end;

  function FModulo( aValue, aDivisor: TSignal): TSignal; inline;
  begin
    Result := aValue - ( Int( aValue / aDivisor) * aDivisor);
  end;

  function  AntiRunAway( aValue: TSignal): TSignal; // inline;
  begin
    if Abs( aValue) < 1e-20
    then Result := 0
    else if aValue > MAX_SIGNAL
    then Result := MAX_SIGNAL
    else if aValue < MIN_SIGNAL
    then Result := MIN_SIGNAL
    else Result := aValue;
  end;

  function  RangeMap( const aValue, anInLow, anInHigh, anOutLow, anOutHigh: TSignal): TSignal; inline;
  begin
    Result := anOutLow + (( aValue - anInLow) / ( anInHigh - anInLow)) * ( anOutHigh - anOutLow);
  end;

  function  AlmostEqual( aValue1, aValue2, aDelta: TSignal): Boolean;  inline;
  begin
    Result := Abs( aValue1 - aValue2) < aDelta;
  end;

  function  AlmostZero( aValue: TSignal): Boolean; inline;
  begin
    Result := Abs( aValue) < DELTA;
  end;

  function  Clip( const aValue: TSignal; const aLow, aHigh: TSignal): TSignal; inline;
  begin
    if aValue < aLow
    then Result := aLow
    else if aValue > aHigh
    then Result := aHigh
    else Result := aValue;
  end;

  function  LogicToSignal( aValue: Boolean): TSignal; inline;
  begin
    if aValue
    then Result := 1
    else Result := -1;
  end;

  function  SignalToLogic( aValue: TSignal): Boolean; inline;
  begin
    Result := aValue > 0;
  end;

  function  SignalToInt( aValue : TSignal): Integer; inline;
  begin
    Result := Round( aValue + 0.4999);
  end;

  function  SignalTodB( aValue : TSignal): TSignal; inline;
  begin
    if aValue <= 1e-10
    then Result := -200
    else Result := 20 * Log10( aValue);
  end;


  function  SignalToNoteNumber( aValue: TSignal): TSignal; inline;
  begin
    Result := 128 * aValue;
  end;

  function  NoteNumberToUnits( aValue: TSignal): TSignal; inline
  begin
    Result := aValue / 128;
  end;

  function  UnitsToFrequency( aValue: TSignal): TSignal; inline;
  // such that note Nr 69 maps to A 440 Hz
  // see: http://www.phys.unsw.edu.au/jw/notes.html
  // freq = ( 2 ^ ( notenr - 69) / 12) * 440
  // Result := Power( 2, ( aValue - 69) / 12) * ReferenceA;
  begin
    Result := Power( 2, ( SignalToNoteNumber( aValue) - MiddleNote) / NotesPerOctave) * ReferenceA;
  end;

  function  UnitsToFrequencyLin( aValue: TSignal): TSignal; inline;
  // such that the interval (0,1) maps to (0 Hz,SYSTEM_RATE / 2 Hz)
  begin
    Result := aValue * SYSTEM_RATE / 2;
  end;

  function  FrequencyToUnits( aValue: TSignal): TSignal; inline;
  // see: http://www.phys.unsw.edu.au/jw/notes.html
  // n  = 12 * log2( f / 440) + 69
  //  Result := Round( 12 * Log2( aValue / ReferenceA)) + 69;
  begin
    Result := NoteNumberToUnits( NotesPerOctave * Log2( aValue / ReferenceA) + MiddleNote);
  end;

  function  NoteNrToFrequency( aValue : TSignal): TSignal; inline;
  begin
    Result := UnitsToFrequency( NoteNumberToUnits( aValue));
  end;

  function  FrequencyToNoteNr( aValue: TSignal): TSignal; inline;
  begin
    Result := SignalToNoteNumber( FrequencyToUnits( aValue));
  end;

  function  FrequencyToPhaseDelta( aSignal: TSignal): TSignal; inline;
  begin
    Result := aSignal / SYSTEM_RATE;
  end;

  function  SampleCountToTime( aCount: TSignal): TSignal; inline;
  begin
    Result := aCount / SYSTEM_RATE;
  end;

  function  TimeToSampleCount( aTime: TSignal): TSignal; inline;
  begin
    Result := aTime * SYSTEM_RATE;
  end;


  // ------------------------------------

  function  SlowUnitsToFrequency( aValue  : TSignal): TSignal; // inline;
  // Fixed conversion for LFOs
  begin
    Result := Power( 2, ( SignalToNoteNumber( aValue) - 69) / 12) * 440; // No need for SlowSignalToNoteNumber here
  end;

  function  SlowNoteNrToFrequency( aValue  : TSignal): TSignal; inline;
  begin
    Result := SlowUnitsToFrequency( NoteNumberToUnits( aValue)); // No need for SlowNoteNrToUnits here
  end;

  function  SlowFrequencyToPhaseDelta( aSignal: TSignal): TSignal; inline;
  begin
    Result := SLOW_DECIMATION * aSignal / SYSTEM_RATE;
  end;

  function  SlowSampleCountToTime( aCount: TSignal): TSignal; inline;
  begin
    Result := aCount / SLOW_RATE;
  end;

  function  SlowTimeToSampleCount( aTime: TSignal): TSignal; inline;
  begin
    Result := aTime * SLOW_RATE;
  end;

  function  NoteNameToUnits( const aNoteName: string; aDefault: TSignal): TSignal;
  // C[#][-]0
  var
    Base  : Integer;
    Sharp : Integer;
    Sign  : Integer;
    N     : Char;
    Shift : Integer;
    p     : Integer;
    T     : string;
    NN    : Integer;
  begin
    Shift :=  0;
    Sharp :=  0;
    Sign  :=  1;
    Base  := -1;
    if Length( aNoteName) > 0
    then begin
      N := UpCase( aNoteName[ 1]);
      case N of
        'C' : Base := 12     ;
        'D' : Base := 12 +  2;
        'E' : Base := 12 +  4;
        'F' : Base := 12 +  5;
        'G' : Base := 12 +  7;
        'A' : Base := 12 +  9;
        'B' : Base := 12 + 11;
        else  Base := -1;
      end;
      if ( Length( aNoteName) > 1) and ( Base > 0)
      then begin
        if aNoteName[ 2] = '#'
        then begin
          Sharp := 1;
          p     := 2;
        end
        else p := 1;
      end
      else p := 0;
      if ( Length( aNoteName) > p) and ( base > 0)
      then begin
        if aNoteName[ p + 1] = '-'
        then begin
          Sign := -1;
          p    := p + 1;
        end;
      end;
      T := Copy( aNoteName, p + 1, Length( aNoteName));
      if T[ Length( T)] = '#'
      then begin
        Sharp := 1;
        T     := Copy( T, 1, Length( T) - 1);
      end;
      Shift := StrToIntDef( T, 0);
    end;
    if Base > 0
    then begin
      NN     := Base + Sharp + Sign * Shift * 12;
      Result := NoteNumberToUnits( NN);
    end
    else Result := aDefault;
  end;

  function  UnitsToNoteName( aValue: TSignal; Compact: Boolean = True): string;
  const
    Notes  : array[ 0 .. 11] Of Char = 'CCDDEFFGGAAB';
    Sharps : array[ 0 .. 11] Of Char = ' # #  # # # ';
  var
    i          : Integer;
    S          : string;
    aNoteValue : Integer;
  begin
    aNoteValue := Round( SignalToNoteNumber( aValue));
    i          := Remainder( aNoteValue, 12);
    S          := Sharps[ i];
    if Compact and ( S = ' ')
    then S := '';
    Result := Format( '%s%s%d', [ Notes[ i], S, ( FlooredDiv( aNoteValue, 12)) - 1], AppLocale);
  end;

  function  IsNoteName( const aNoteName: string): Boolean;
  // C[#][-]0[#]
  var
    p : Integer;
    i : Integer;
  begin
    Result := False;
    if Length( aNoteName) > 0
    then begin
      Result := CharInSet( aNoteName[ 1], ['a' .. 'g', 'A' .. 'G']);
      if Result and ( Length( aNoteName) > 1)
      then begin
        if aNoteName[ 2] = '#'
        then p := 2
        else p := 1;
      end
      else p := 0;

      if Result and ( Length( aNoteName) > p) and ( aNoteName[ p + 1] = '-')
      then p := p + 1;

      if Result
      then begin
        for i := p + 1 to Length( aNoteName)
        do begin
          if not CharInset( aNoteName[ i], [ '0' .. '9'])
          then begin
            if ( aNoteName[ i] = '#') and ( i = Length( aNoteName))
            then Break
            else begin
              Result := False;
              Break;
            end;
          end;
        end;
      end;
    end;
  end;

  procedure ParsePrefix( const aPath: string; var aPrefix, aRest: string);
  // NOTE: this code needs be compiled with "String Format Checking" on,
  //       otherwise it will generate a run time error.
  //       ok .. this will only occur when aPath and aRest refer to the same
  //       underlying string .. as in ParsePrefix( aRest, aPrefix, aRest);
  var
    i  : Integer;
  begin
    i       := 1;
    aPrefix := '';
    aRest   := '';
    while i <= Length( aPath)
    do begin
      if aPath[ i] = '/'
      then begin
        aPrefix := Copy( aPath, 1, i - 1 );                    // Either on this line
        aRest := Copy( aPath, i + 1, Length( aPath));          // Or this one .. not quite clear.
        Break;
      end
      else Inc( i);
    end;
  end;

  function  ParsePostfix( const aPath: string): string;
  var
    i : Integer;
  begin
    Result := '';
    for i := Length( aPath) downto 1
    do begin
      if aPath[ i] = '_'
      then begin
        Result := Copy( aPath, i + 1, Length( aPath));
        Break;
      end
    end;
  end;

  function  ApplyUnitsToSignal( var aValue: TSignal; const aUnits: string): Boolean;
  begin
    Result := True;
    if      SameText( aUnits, '')
    then // nothing
    else if SameText( aUnits, 'Hz')
    then aValue := FrequencyToUnits( aValue)
    else if SameText( aUnits, 'kHz')
    then aValue := FrequencyToUnits( 1000 * aValue)
    else if SameText( aUnits, 'mHz')
    then aValue := FrequencyToUnits( aValue / 1000)
    else if SameText( aUnits, 'bpm')
    then aValue := FrequencyToUnits( aValue / 60.0)
    else if SameText( aUnits, 'degrees') or SameText( aUnits, 'deg')
    then aValue := aValue / 360
    else if SameText( aUnits, 'radians') or SameText( aUnits, 'rad')
    then aValue := aValue / ( PI_2)
    else if SameText( aUnits, 'percent') or SameText( aUnits, '%')
    then aValue := aValue / 100
    else if SameText( aUnits, 'ms')
    then aValue := aValue / 1000
    else if SameText( aUnits, 's')
    then // nothing
    else if SameText( aUnits, 'nn')
    then aValue := NoteNumberToUnits( aValue)
    else if SameText( aUnits, 'qsteps') or SameText( aUnits, 'q')
    then aValue := aValue / 1000
    else Result := False;
  end;

  function  StripWhiteSpace( const S: string): string;
  var
    C: Char;
  begin
    Result := '';
    for C in S
    do begin
      if not CharInSet( C, WhiteSpace)
      then Result := Result + C;
    end;
  end;

  function  ParseTextValue( const aString: string; var aSignal: TSignal): Boolean;
  var
    aParts  : TStringList;
    S       : string;
    aUnits  : string;
  begin
    Result := True;
    aParts := Explode( aString, ' ');
    try
      aUnits := '';
      if aParts.Count > 0
      then begin
        S := Trim( aParts[ 0]);
        if IsNoteName( S)
        then aSignal := NoteNameToUnits( S, UNDEFINED)
        else aSignal := StrToFloatDef  ( S, UNDEFINED);
      end
      else Result := False;
      if Result and ( aParts.Count > 1)
      then begin
        aUnits := Trim( aParts[ 1]);
        Result := ApplyUnitsToSignal( aSignal, aUnits);
      end;
    finally
      aParts.Free;
    end;
  end;

  function  ParseTextSequencerValue( const aValue: string): TSignals;
  var
    aStrings : TStringList;
    i        : Integer;
    aSignal  : TSignal;
  begin
    Result   := TSignals.Create( 0);
    aStrings := Explode( StripWhiteSpace( aValue), ',');
    try
      for i := 0 to aStrings.Count - 1
      do begin
        if ParseTextValue( aStrings[ i], aSignal)
        then Result.AddSignal( aSignal);
      end;
    finally
      aStrings.Free;
    end;
  end;

  function  ExpAttack( aTime, aDuration, aStartValue, anEndValue: TSignal): TSignal;
  // y = 2 ^ x - 1  [0,1] -> [0,1]
  var
    Unscaled: TSignal;
  begin
    Assert( aDuration > 0);
    aTime    := Clip( aTime, 0, aDuration) / aDuration;
    Unscaled := ( Power( 2, 10 * aTime) - 1) / 1023;
    Result   := ( 1 - Unscaled) * aStartValue + Unscaled * anEndValue;
  end;

  function  LinAttack( aTime, aDuration, aStartValue, anEndValue: TSignal): TSignal;
  // y = x  [0,1] -> [0,1]
  var
    Unscaled: TSignal;
  begin
    Assert( aDuration > 0);
    aTime    := Clip( aTime, 0, aDuration) / aDuration;
    Unscaled := aTime;
    Result   := ( 1 - Unscaled) * aStartValue + Unscaled * anEndValue;
  end;

  function  LogAttack( aTime, aDuration, aStartValue, anEndValue: TSignal): TSignal;
  // y = 2 - 2 ^ ( 1 - x)  [0,1] -> [0,1]
  var
    Unscaled: TSignal;
  begin
    Assert( aDuration > 0);
    aTime    := Clip( aTime, 0, aDuration) / aDuration;
    Unscaled := ( 1024 - Power( 2, 10 - 10 * aTime)) / 1023;
    Result   := ( 1 - Unscaled) * aStartValue + Unscaled * anEndValue;
  end;

  function  ExpDecay( aTime, aDuration, aStartValue, anEndValue: TSignal): TSignal;
  // y = 2 ^ ( 1 - x) - 1  [0,1] -> [0,1]
  var
    Unscaled: TSignal;
  begin
    Assert( aDuration > 0);
    aTime    := Clip( aTime, 0, aDuration) / aDuration;
    Unscaled := ( Power( 2, 10 - 10 * aTime) - 1) / 1023;
    Result   := ( 1 - Unscaled) * anEndValue + Unscaled * aStartValue;
  end;

  function  LinDecay( aTime, aDuration, aStartValue, anEndValue: TSignal): TSignal;
  // y = 1 - x  [0,1] -> [0,1]
  var
    Unscaled: TSignal;
  begin
    Assert( aDuration > 0);
    aTime    := Clip( aTime, 0, aDuration) / aDuration;
    Unscaled := 1 - aTime;
    Result   := ( 1 - Unscaled) * anEndValue + Unscaled * aStartValue;
  end;

  function  LogDecay( aTime, aDuration, aStartValue, anEndValue: TSignal): TSignal;
  // y = 2 - 2 ^ x  [0,1] -> [0,1]
  var
    Unscaled: TSignal;
  begin
    Assert( aDuration > 0);
    aTime    := Clip( aTime, 0, aDuration) / aDuration;
    Unscaled := ( 1024 - Power( 2, 10 * aTime)) / 1023;
    Result   := ( 1 - Unscaled) * anEndValue + Unscaled * aStartValue;
  end;

  function  MakeMidiNoteOff( aVelocity, aNote, aChannel: Byte): TMidiMessage;
  begin
    Result.Command := CNoteOff;
    Result.Channel := aChannel;
    SetLength( Result.Data, 2);
    Result.Data[ 0] := aNote;
    Result.Data[ 1] := aVelocity;
  end;

  function  MakeMidiNoteOn( aVelocity, aNote, aChannel: Byte): TMidiMessage;
  begin
    Result.Command := CNoteOn;
    Result.Channel := aChannel;
    SetLength( Result.Data, 2);
    Result.Data[ 0] := aNote;
    Result.Data[ 1] := aVelocity;
  end;

  function  MakeMidiKeyPressure( aPressure, aNote, aChannel: Byte): TMidiMessage;
  begin
    Result.Command := CKeyPressure;
    Result.Channel := aChannel;
    SetLength( Result.Data, 2);
    Result.Data[ 0] := aNote;
    Result.Data[ 1] := aPressure;
  end;

  function  MakeMidiControlChange( aValue, aController, aChannel: Byte): TMidiMessage;
  begin
    Result.Command := CControlChange;
    Result.Channel := aChannel;
    SetLength( Result.Data, 2);
    Result.Data[ 0] := aController;
    Result.Data[ 1] := aValue;
  end;

  function  MakeMidiProgramChange( aProgram, aChannel: Byte): TMidiMessage;
  begin
    Result.Command := CProgramChange;
    Result.Channel := aChannel;
    SetLength( Result.Data, 1);
    Result.Data[ 0] := aProgram;
  end;

  function  MakeMidiChannelPressure( aValue, aChannel: Byte): TMidiMessage;
  begin
    Result.Command := CChannelPressure;
    Result.Channel := aChannel;
    SetLength( Result.Data, 1);
    Result.Data[ 0] := aValue;
  end;

  function  MakeMidiPitchBend( aData: Word; aChannel: Byte): TMidiMessage;
  begin
    Result.Command := CPitchBend;
    Result.Channel := aChannel;
    SetLength( Result.Data, 1);
    Result.Data[ 0] := aData;
  end;

  function  MakeMidiSongPosition( aData: Word): TMidiMessage;
  begin
    Result.Command := SSongPosition;
    Result.Channel := 0;
    SetLength( Result.Data, 1);
    Result.Data[ 0] := aData;
  end;

  function  MakeMidiSongSelect( aSong: Byte): TMidiMessage;
  begin
    Result.Command := SSongSelect;
    Result.Channel := 0;
    SetLength( Result.Data, 1);
    Result.Data[ 0] := aSong;
  end;

  function  MakeMidiTuneRequest: TMidiMessage;
  begin
    Result.Command := STuneRequest;
    Result.Channel := 0;
    SetLength( Result.Data, 0);
  end;

  function  MakeMidiTimingClock: TMidiMessage;
  begin
    Result.Command := STimingClock;
    Result.Channel := 0;
    SetLength( Result.Data, 0);
  end;

  function  MakeMidiStart: TMidiMessage;
  begin
    Result.Command := SStart;
    Result.Channel := 0;
    SetLength( Result.Data, 0);
  end;

  function  MakeMidiContinue: TMidiMessage;
  begin
    Result.Command := SContinue;
    Result.Channel := 0;
    SetLength( Result.Data, 0);
  end;

  function  MakeMidiStop: TMidiMessage;
  begin
    Result.Command := SStop;
    Result.Channel := 0;
    SetLength( Result.Data, 0);
  end;

  function  MakeMidiActiveSense: TMidiMessage;
  begin
    Result.Command := SActiveSense;
    Result.Channel := 0;
    SetLength( Result.Data, 0);
  end;

  function  MakeMidiReset: TMidiMessage;
  begin
    Result.Command := SReset;
    Result.Channel := 0;
    SetLength( Result.Data, 0);
  end;

  function  MakeMidiSysExData( const aData: TBytes): TMidiMessage;
  var
    i : Integer;
  begin
    Result.Command := SSysEx;
    Result.Channel := 0;
    SetLength( Result.Data, Length( aData));
    for i := 0 to Length( Result.Data) - 1
    do Result.Data[ i] := aData[ i];
  end;

  procedure MidiMessageToBytes( const aMsg: TMidiMessage; var aData: TBytes);
  begin
    SetLength( aData, 0);

    case aMsg.Command of

      CNoteOff,          // = $80; // Length = 3;
      CNoteOn,           // = $90; // Length = 3;
      CKeyPressure,      // = $a0; // Length = 3;
      CControlChange :   // = $b0; // Length = 3;
        begin
          if Length( aMsg.Data) = 2
          then begin
            SetLength( aData, 3);
            aData[ 0] := ( aMsg.Command and $f0) or ( aMsg.Channel and $0f);
            aData[ 1] := aMsg.Data[ 0];
            aData[ 2] := aMsg.Data[ 1];
          end;
        end;

      CProgramChange,    // = $c0; // Length = 2;
      CChannelPressure : // = $d0; // Length = 2;
        begin
          if Length( aMsg.Data) = 2
          then begin
            SetLength( aData, 2);
            aData[ 0] := ( aMsg.Command and $f0) or ( aMsg.Channel and $0f);
            aData[ 1] := aMsg.Data[ 0];
          end;
        end;

      CPitchBend :       // = $e0; // Length = 3;
        begin
          if Length( aMsg.Data) = 2
          then begin
            SetLength( aData, 3);
            aData[ 0] := ( aMsg.Command  and $00f0) or ( aMsg.Channel and $000f);
            aData[ 1] :=   aMsg.Data[ 0] and $007f;
            aData[ 2] := ( aMsg.Data[ 0] and $3f80) shr 7;
          end;
        end;

//      CSysEx :          // = $f0; // Length = n;
//        begin
//        end;

    end;
  end;

  function  MidiMessageToLog( const aMsg: TMidiMessage): string;

    function CmdToStr( aCmd: Byte): string;
    begin
      case aCmd of
        CNoteOff         : Result := 'note off';         // = $80; // Length = 3;
        CNoteOn          : Result := 'note on';          // = $90; // Length = 3;
        CKeyPressure     : Result := 'key pressure';     // = $a0; // Length = 3;
        CControlChange   : Result := 'control change';   // = $b0; // Length = 3;
        CProgramChange   : Result := 'program change';   // = $c0; // Length = 2;
        CChannelPressure : Result := 'channel pressure'; // = $d0; // Length = 2;
        CPitchBend       : Result := 'pitch bend';       // = $e0; // Length = 2;
        CSysEx           : Result := 'sysex';            // = $f0; // Length = n;
      end;
    end;

  var
    i : Integer;
  begin
    Result := Format( 'ch %d cmd %.2x [%s] ', [ aMsg.Channel + 1, aMsg.Command, CmdToStr( aMsg.Command)]);
    for i := 0 to Length( aMsg.Data) - 1
    do Result := Result + Format( '%.2x ', [ aMSg.Data[ i]]);
    SetLength( Result, Length( Result) - 1);
  end;


  procedure InitializeUnit;
  begin
    ReferenceA     := 440.0;
    NotesPerOctave := 12.0;
    MiddleNote     := 69;
    FModuleTypes   := TModuleTypes.Create;
  end;

  procedure FinalizeUnit;
  begin
    FreeAndNil( FModuleTypes);
  end;




{ ========
  TModuleTypes = class( TStringList)
  public
}

    constructor TModuleTypes.Create;
    begin
      inherited;
      Duplicates := dupError;
    end;

    procedure   TModuleTypes.RegisterType( aModuleType: TModuleClass);
    begin
      if Assigned( aModuleType)
      then AddObject( aModuleType.ClassName, TObject( aModuleType));
    end;

    function    TModuleTypes.CreateModule( aParent: TSynthPatch; const aTypeName, aName: string; anInputCount: Integer): TMod;
    var
      i : Integer;
      T: TModuleClass;
    begin
      Result := nil;
      i := IndexOf( aTypeName);
      if i >= 0
      then begin
        T := TModuleClass( Objects[ i]);
        Result := T.Create( aParent, aName, anInputCount);
      end;
    end;


  procedure RegisterModuleType( aModuleType: TModuleClass);
  begin
    FModuleTypes.RegisterType( aModuleType);
  end;

  procedure RegisterModuleTypes( aModuleTypes: array of TModuleClass);
  var
    M: TModuleClass;
  begin
    for M in aModuleTypes
    do RegisterModuleType( M);
  end;



{ ========
  TSignal          = Double;
  TSignalArray     = array of TSignal;
  TConnection      = class;
  TConnectionArray = array of TConnection;
  TMod             = class;
  TModuleArray     = array of TModule;

  TSignals = class
  private
    FSignals: TSignalArray;
  public
    property    Count                    : Integer read GetCount;
    property    Signal[ anIndex: Integer]: TSignal read GetSignal write SetSignal;
  private
}

    function    TSignals.GetCount: Integer;
    begin
      Result := Length( FSignals);
    end;

    function    TSignals.GetSignal( anIndex: Integer): TSignal;
    begin
      Result := FSignals[ anIndex];
    end;

    procedure   TSignals.SetSignal( anIndex: Integer; aValue: TSignal);
    begin
      FSignals[ anIndex] := aValue;
    end;

    function    TSignals.GetInterpolate( anIndex: TSignal): TSignal;
    var
      Lower   : Integer;
      Higher  : Integer;
      DLower  : TSignal;
      DHigher : TSignal;
    begin
      Lower  := Floor( anIndex);
      DLower := anIndex - Lower;
      if Lower < 0
      then Lower := Count - 1;
      Higher := Ceil( anIndex);
      if Higher = Lower
      then Higher := Lower + 1;
      DHigher := Higher - anIndex;
      if Higher > Count - 1
      then Higher := 0;
      Result := DHigher * Signal[ Lower] + DLower * Signal[ Higher];
    end;

//  public

    constructor TSignals.Create( aCount: Integer); // overload;
    begin
      inherited Create;
      SetLength( FSignals, aCount);
    end;

    destructor  TSignals.Destroy; // override;
    begin
      Clear;
      inherited;
    end;

    procedure   TSignals.Clear;
    begin
      SetLength( FSignals, 0);
    end;

    procedure   TSignals.AddSignal( aValue: TSignal);
    begin
      SetLength( FSignals, Count + 1);
      FSignals[ Count - 1] := aValue;
    end;

    procedure   TSignals.SetSignals( aStartIndex: Integer; const aValue: TSignals);
    var
      i : Integer;
    begin
      if Assigned( aValue)
      then begin
        SetLength( FSignals, aStartIndex + aValue.Count);
        for i := aStartIndex to Count - 1
        do Signal[ i] := aValue.Signal[ i - aStartIndex];
      end;
    end;

    function    TSignals.Convolve( aValue: TSignal): TSignal;
    var
      i : Integer;
    begin
      Result := 0;
      for i := 0 to Count - 1
      do Result := Result + aValue * Signal[ i];
    end;


{ ========
  TSignalPair = record
    Left  : TSignal;
    Right : TSignal;
  end;

  TSignalPairs = array of TSignalPair;

  TSignalQueue = class
  private
    FSignals    : TSignalPairs;
    FInp        : Integer;
    FOutp       : Integer;
    FFreeCount  : Integer;
    FFillCount  : Integer;
    FOverflows  : Integer;
    FUnderflows : Integer;
    FLastValue  : TSignalPair;
  private
}

    function    TSignalQueue.GetOverflows: Integer;
    begin
      Result := FOverflows;
      FOverflows := FOverflows - Result;
    end;

    function    TSignalQueue.GetUnderflows: Integer;
    begin
      Result := FUnderflows;
      FUnderflows := FUnderflows - Result;
    end;

//  public

    constructor TSignalQueue.Create( aSize: Integer);
    begin
      inherited Create;
      SetLength( FSignals, aSize);
      Clear;
    end;

    destructor  TSignalQueue.Destroy; // override;
    begin
      Clear;
      FFreeCount := 0;
      FFillCount := 0;
      inherited;
    end;

    procedure   TSignalQueue.Clear;
    begin
      FFillCount       := 0;
      FFreeCount       := 0;
      FInp             := 0;
      FOutp            := 0;
      FUnderflows      := 0;
      FOverflows       := 0;
      FLastValue.Left  := 0;
      FLastValue.Right := 0;
      FFreeCount       := Length( FSignals);
    end;

    procedure   TSignalQueue.Append( aLeft, aRight: TSignal); // overload;
    var
      aPair: TSignalPair;
    begin
      aPair.Left  := aLeft;
      aPair.Right := aRight;
      Append( aPair);
    end;

    procedure   TSignalQueue.Append( aPair: TSignalPair); // overload;
    begin
      if FFreeCount > 0
      then begin
        Dec( FFreeCount);
        FSignals[ FInp] := aPair;
        FInp := ( FInp + 1) mod Length( FSignals);
        Inc( FFillCount);
      end
      else {$Q-} Inc( FOverflows); {$Q+}
    end;

    function    TSignalQueue.GetData: TSignalPair;
    begin
      if FFillCount > 0
      then begin
        Dec( FFillCount);
        FLastValue := FSignals[ FOutp];
        FOutp      := ( FOutp + 1) mod Length( FSignals);
        Inc( FFreeCount);
      end
      else {$Q-} Inc( FUnderflows); {$Q+}
      Result := FLastValue;
    end;


{ ========
  TConnection = class
  private
    FPeer      : TPatch;
    FIndex     : Integer;
    FSrcModule : Integer;
    FSrcPin    : Integer;
    FDstModule : Integer;
    FDstPin    : Integer;
  public
    property    SrcModule: Integer read FSrcModule;
    property    SrcPin   : Integer read FSrcPin;
    property    DstModule: Integer read FDstModule;
    property    DstPin   : Integer read FDstPin;
  private
}

    procedure   TConnection.Connect( const aPeer: TSynthPatch; anIndex, aSrcModule, aSrcPin, aDstModule, aDstPin: Integer);
    begin
      Peer      := aPeer;
      Index     := anIndex;
      SrcModule := aSrcModule;
      SrcPin    := aSrcPin;
      DstModule := aDstModule;
      DstPin    := aDstPin;
    end;

//  public

    procedure   TConnection.Save( anIndent: Integer; var S: string);
    var
      aSrcModName : string;
      aSrcPinName : string;
      aDstModName : string;
      aDstPinName : string;
      T           : string;
    begin
      if Assigned( Peer)
      then begin
        aSrcModName := Peer.ModuleName   ( SrcModule);
        aSrcPinName := Peer.OutputPinName( SrcModule, SrcPin);
        aDstModName := Peer.ModuleName   ( DstModule);
        aDstPinName := Peer.InputPinName ( DstModule, DstPin);

        T := Format( '%sconnect( %s %s %s %s)' + sLineBreak, [ Indent( anIndent), aSrcModName, aSrcPinName, aDstModName, aDstPinName], AppLocale);
        S := S + T
      end;
    end;

    function    TConnection.IsSameConnection( aConnection: TConnection): Boolean;
    begin
      Result := False;
      if Assigned( aConnection)
      then
        Result :=
          ( SrcModule = aConnection.SrcModule) and
          ( SrcPin    = aConnection.SrcPin   ) and
          ( DstModule = aConnection.DstModule) and
          ( DstPin    = aConnection.DstPin   );
    end;


{ ========
  TConnections = class
  private
    FConnections: TConnectionArray;
  public
    property    Count: Integer read GetCount;
    property    Connection[ anIndex: Integer]: TConnection read GetConnection write SetConnection;
  private
}

    function    TConnections.GetCount: Integer;
    begin
      Result := Length( FConnections);
    end;

    function    TConnections.GetConnection( anIndex: Integer): TConnection;
    begin
      Result := FConnections[ anIndex]
    end;

    procedure   TConnections.SetConnection( anIndex: Integer; const aValue: TConnection);
    begin
      FConnections[ anIndex] := aValue;
    end;

//  public

    constructor TConnections.Create( aPeer: TSynthPatch);
    begin
      inherited Create;
      FPeer := aPeer;
    end;

    destructor  TConnections.Destroy; // override;
    begin
      Clear;
      inherited;
    end;

    procedure   TConnections.Save( anIndent: Integer; const aName: string; var S: string);
    var
      i : Integer;
    begin
      S := S + Indent( anIndent) + aName + '(' + sLineBreak;
      for i := 0 to Count - 1
      do Connection[ i].Save( anIndent + 1, S);
      S := S + Indent( anIndent) + ')' + sLineBreak;
    end;

    procedure   TConnections.Clear;
    begin
      while Count > 0
      do RemoveConnection( Count - 1);
    end;

    procedure   TConnections.AddConnection( const aValue: TConnection); // overload;
    begin
      SetLength( FConnections, Count + 1);
      FConnections[ Count - 1] := aValue;
    end;

    procedure   TConnections.AddConnection( aSrcModule, aSrcPin, aDstModule, aDstPin: Integer);  // overload;
    var
      aConnection: TConnection;
    begin
      aConnection := TConnection.Create;
      aConnection.Connect( FPeer, Count, aSrcModule, aSrcPin, aDstModule, aDstPin);
      AddConnection( aConnection);
    end;

    procedure   TConnections.RemoveConnection( anIndex: Integer); // overload;
    var
      aConnection: TConnection;
      i          : Integer;
    begin
      aConnection := Connection[ anIndex];
      if Assigned( aConnection)
      then begin
        for i := anIndex to Count - 2
        do Connection[ i] := Connection[ i + 1];
        SetLength( FConnections, Count - 1);
        aConnection.Free;
      end;
    end;

    procedure   TConnections.RemoveConnectionsFor( aModuleIndex: Integer);
    var
      i : Integer;
      C: TConnection;
    begin
      for i := Count - 1 downto 0
      do begin
        C := Connection[ i];
        if ( C.SrcModule = aModuleIndex) or ( C.DstModule = aModuleIndex)
        then RemoveConnection( i);
      end;
    end;

    procedure   TConnections.PropagateSignals( const aModules: TModuleArray);
    var
      i : Integer;
    begin
      for i := 0 to Count - 1
      do begin
        with Connection[ i]
        do aModules[ DstModule].FInCache[ DstPin] := aModules[ SrcModule].FOutCache[ SrcPin];
      end;
    end;

    function    TConnections.IsSameConnections( const aConnections: TConnections): Boolean;
    var
      i : Integer;
    begin
      Result := False;
      if Assigned( aConnections) and ( Count = aConnections.Count)
      then begin
        Result := True;
        for i := 0 to Count - 1
        do begin
          if not Connection[ i].IsSameConnection( aConnections.Connection[ i])
          then begin
            Result := False;
            Break;
          end;
        end;
      end;
    end;


{ ========
  TMod = class
  private
    FParent      : TSynthPatch;
    FName        : string;
    FInputs      : TSignals;
    FOutputs     : TSignals;
    FTickLocks   : Integer;
    FSamples     : Int64;
    FClocks      : Int64;
    FTotalClocks : Int64;
  protected
    FIsSlow   : Boolean;
    FIsFast   : Boolean;
    FIsMidi   : Boolean;
    FInCache  : TSignalArray;
    FOutCache : TSignalArray;
  public
    property    Name                             : string   read FName                  write SetName;
    property    TotalClocks                      : Int64    read FTotalClocks;
    property    Samples                          : Int64    read FSamples;
    property    InputCount                       : Integer  read GetInputCount;
    property    OutputCount                      : Integer  read GetOutputCount;
    property    Output[ anIndex: Integer]        : TSignal  read GetOutput;
  public
    property    ParamValue         [ const aName: string] : TSignal                     write SetParamValue;
    property    InternalValue      [ const aName: string] : TSignal                     write SetInternalValue;
    property    InternalStringValue[ const aName: string] : string                      write SetInternalStringValue;
    property    StringValue        [ const aName: string] : string  read GetStringValue write SetStringValue;
  private
}

    function    TMod.AssertUniqueName( const aName: string): string; // virtual;
    begin
      if Assigned( FParent)
      then Result := FParent.AssertUniqueName( aName)
      else Result := aName;
    end;

//  private

    procedure   TMod.SetName( const aValue: string);
    begin
      if aValue <> FName
      then FName := AssertUniqueName( aValue);
    end;

    function    TMod.GetInputCount: Integer;
    begin
      Result := Length( FInCache);
    end;

    function    TMod.GetOutputCount: Integer;
    begin
      Result := Length( FOutCache);
    end;

    function    TMod.GetOutput( anIndex: Integer): TSignal;
    begin
      Result := FOutCache[ anIndex];
    end;

    procedure   TMod.SetCommonValue( anIndex: Integer; aValue: TSignal); // virtual;
    begin
      if anIndex >= 0
      then FInCache[ anIndex] := aValue;
    end;

    procedure   TMod.SetParamValue( const aName: string; aValue: TSignal); // virtual;
    var
      anIndex: Integer;
    begin
      anIndex := FindInput( aName);
      SetCommonValue( anIndex, aValue);
    end;

    procedure   TMod.SetInternalValue( const aName: string; aValue: TSignal); // virtual;
    var
      anIndex: Integer;
    begin
      anIndex := FindInput( aName);
      if anIndex >= 0
      then SetCommonValue( anIndex, aValue)
      else SetInternal( aName, aValue);
    end;

    procedure   TMod.SetInternalStringValue( const aName, aValue: string);
    var
      aSignal : TSignal;
    begin
      if ParseTextValue( aName, aSignal)
      then InternalValue[ aName] := aSignal
      else raise EUnitsUndefined.CreateFmt( 'Unparsable value ''%s''', [ aName])
    end;

//  protected

    procedure   TMod.SetStringValue( const aName, aValue: string); // virtual;
    begin
      // Nothing here.
    end;

    function    TMod.GetStringValue( const aName: string): string; // virtual;
    begin
      Result := '';
    end;

    procedure   TMod.FixOutputs; // virtual;
    var
      i : Integer;
    begin
      i := 0;
      while not SameText( OutputName( i), NO_OUTPUT)
      do begin
        Inc( i);
        if i > 1000
        then raise EOutputNotFound.Create( 'are there really over 1000 outputs? (probably a GetOutputName() is broken and never returns NO_OUTPUT)');
      end;
      SetOutputCount( i);
    end;

    procedure   TMod.SetOutputCount( aCount: Integer);
    begin
      FOutputs.Clear;
      FOutCache := FOutputs.FSignals;
      while OutputCount < aCount
      do begin
        FOutputs.AddSignal( 0.0);
        FOutCache := FOutputs.FSignals;
      end;
    end;

    procedure   TMod.SetInputs( aStartIndex: Integer; const aValue: TSignals);
    begin
      FInputs.SetSignals( aStartIndex, aValue);
      FInCache := FInputs.FSignals;
    end;

    procedure   TMod.HandleParam( const aControl: string; aValue: TSignal; IsFinal: Boolean); // virtual;
    // Handle a parameter set from the editor patch
    begin
      if IsFinal
      then InternalValue[ aControl] := aValue
      else ParamValue   [ aControl] := aValue;
    end;

    procedure   TMod.ReleaseTickLock;
    begin
      if FTickLocks > 0
      then Dec( FTickLocks);
    end;

    function    TMod.IsTickLocked: Boolean;
    begin
      Result := FTickLocks > 0;
    end;

    function    TMod.GrabTickLock: Boolean;
    begin
      Result := FTickLocks = 0;
      if Result
      then Inc( FTickLocks);
    end;

//  public

    constructor TMod.Create( aParent: TSynthPatch; const aName: string; anInputCount: Integer); // virtual;
    begin
      inherited Create;
      FIsSlow      := False;
      FIsFast      := True;
      FIsMidi      := False;
      FTotalClocks := 0;
      FParent      := aParent;
      Name         := aName;
      FInputs      := TSignals.Create( anInputCount);
      FInCache     := FInputs.FSignals;
      FOutputs     := TSignals.Create( 0);
      FOutCache    := FOutputs.FSignals;
      FixOutputs;
      if anInputCount > 0
      then SetDefaults;
    end;

    destructor  TMod.Destroy; // override;
    begin
      FParent := nil;
      FreeAndNil( FOutputs);
      FreeAndNil( FInputs);
      inherited;
    end;

    procedure   TMod.SetDefaults; // virtual;
    begin
    end;

    function    TMod.SensibleSignal( aValue: TSignal): string;
    begin
      Result := FloatToStrF( aValue, ffGeneral, 15, 0, AppLocale);
    end;

    procedure   TMod.SaveKeyValue( anIndent: Integer; const aName: string; aValue: TSignal; var S: string);
    var
      aBaseName : string;
      i         : Integer;
      Output    : string;
    begin
      aBaseName := '';
      for i := 1 to Length( aName)
      do begin
        if CharInSet( aName[ i], IdentifierStarters)
        then abaseName := abaseName + aName[ i]
        else Break;
      end;
      if SameText( aBaseName, 'frequency') or SameText( aBaseName, 'fm')
      then Output := Format( '''%s Hz''', [ SensibleSignal( UnitsToFrequency( aValue))], AppLocale)
      else if SameText( aBaseName, 'phase')
      then Output := Format( '''%s degrees''', [ SensibleSignal( 360 * aValue)], AppLocale)
      else if SameText( aBaseName, 'pwm') or SameText( aBaseName, 'delay')
      then Output := Format( '''%s %%''', [ SensibleSignal( aValue * 100)], AppLocale)
      else if SameText( aBaseName, 'length') or SameText( aBaseName, 'attack') or SameText( aBaseName, 'decay')or SameText( aBaseName, 'release')
      then Output := Format( '''%s s''', [ SensibleSignal( aValue)], AppLocale)
      else if SameText( aBaseName, 'qsteps')
      then Output := Format( '''%s qsteps''', [ SensibleSignal( 1000 * aValue)], AppLocale)
      else if SameText( aBaseName, 'step')
      then Output := UnitsToNoteName( aValue)
      else Output := Format( '%s', [ SensibleSignal( aValue)], AppLocale);
      S := S + Format( Indent( anIndent + 1) + '%s %s' + sLineBreak, [ aName, Output], AppLocale);
    end;

    procedure   TMod.SaveInputValues( anIndent: Integer; var S: string);
    var
      i : Integer;
      N : string;
      p : Integer;
    begin
      p := 1;
      S := S + Indent( anIndent) + 'inputs(' + sLineBreak;
      for i := 0 to InputCount - 1
      do begin
        N := InputName( i);
        if SameText( N, NO_INPUT)
        then begin
          N := Format( 'in%d', [ p], AppLocale);
          Inc( p);
        end;
        SaveKeyValue( anIndent, N, FInCache[ i], S);
      end;
      S := S + Indent( anIndent) + ')' + sLineBreak;
    end;

    procedure   TMod.SaveDetails( anIndent: Integer; var S: string); // virtual;
    begin
      S := S + Format( Indent( anIndent) + '%s' + sLineBreak, [ ClassName ], AppLocale);
      S := S + Format( Indent( anIndent) + '%s' + sLineBreak, [ Name      ], AppLocale);
      SaveInternals  ( anIndent, S);
      SaveInputValues( anIndent, S);
    end;

    procedure   TMod.SaveInternals( anIndent: Integer; var S: string); // virtual;
    var
      aName: string;
      i    : Integer;
    begin
      aName := InternalName( 0);
      if not SameText( aName, NO_INTERNAL)
      then begin
        S := S + Indent( anIndent) + 'internals(' + sLineBreak;
        i := 0;
        while not SameText( aName, NO_INTERNAL)
        do begin
          SaveKeyValue( anIndent, aName, GetInternal( aName), S);
          Inc( i);
          if i > 1000
          then raise EInternalNotFound.Create( 'Are there really over 1000 internals??');
          aName := InternalName( i);
        end;
        S := S + Indent( anIndent) + ')' + sLineBreak;
      end;
    end;

    procedure   TMod.SaveStrings( anIndent: Integer; var S: string); // virtual;
    var
      aName: string;
      i    : Integer;
    begin
      aName := StringName( 0);
      if not SameText( aName, NO_STRING)
      then begin
        S := S + Indent( anIndent) + 'strings(' + sLineBreak;
        i := 0;
        while not SameText( aName, NO_STRING)
        do begin
          S := S + Format( Indent( anIndent + 1) + '%s %s' + sLineBreak, [ aName, StringValue[ aName]], AppLocale);
          Inc( i);
          if i > 1000
          then raise EInternalNotFound.Create( 'Are there really over 1000 strings??');
          aName := InternalName( i);
        end;
        S := S + Indent( anIndent) + ')' + sLineBreak;
      end;
    end;

    procedure   TMod.SetInternal( const aName: string; aValue: TSignal); // virtual;
    begin
      // Internals is an option not present here.
    end;

    function    TMod.GetInternal( const aName: string): TSignal; // virtual;
    begin
      // Internals is an option not present here.
      Result := UNDEFINED;
    end;

    procedure   TMod.Save( anIndent: Integer; var S: string);
    begin
      S := S + Indent( anIndent) + 'module(' + sLineBreak;
      SaveDetails  ( anIndent + 1, S);
      S := S + Indent( anIndent) + ')' + sLineBreak;
    end;

    procedure   TMod.Reset; // virtual;
    begin
    end;

    procedure   TMod.Clear; // virtual;
    begin
    end;

    function    TMod.FindInput( const aPinName: string): Integer; // virtual;
    var
      i : Integer;
    begin
      Result := UNDEF_POMOC;
      for i := 0 to InputCount - 1
      do begin
        if SameText( aPinName, InputName( i))
        then begin
          Result := i;
          Break;
        end;
      end;
    end;

    function    TMod.FindOutput( const aPinName: string): Integer; // virtual;
    var
      i : Integer;
    begin
      Result := UNDEF_POMOC;
      for i := 0 to OutputCount - 1
      do begin
        if SameText( aPinName, OutputName( i))
        then begin
          Result := i;
          Break;
        end;
      end;
    end;

    function    TMod.InputName( anIndex: Integer): string; // virtual;
    begin
      Result := NO_INPUT;
    end;

    function    TMod.InternalName( anIndex: Integer): string; // virtual;
    begin
      Result := NO_INTERNAL;
    end;

    function    TMod.StringName( anIndex: Integer): string; // virtual;
    begin
      Result := NO_STRING;
    end;

    function    TMod.OutputName( anIndex: Integer): string; // virtual;
    begin
      Result := NO_OUTPUT;
    end;

    procedure   TMod.PreTick;
    begin
      {$Q-}
      Inc( FSamples);
      FClocks := GetCpuClockCycleCount;
      {$Q+}
    end;

    procedure   TMod.PostTick;
    begin
      {$Q-}
      FClocks      := GetCpuClockCycleCount - FClocks;
      FTotalClocks := FTotalClocks + FClocks;
      {$Q+}
    end;

    procedure   TMod.Tick;
    begin
      PreTick;
      DoTick;
      PostTick;
    end;

    Procedure   TMod.SlowTick;
    begin
      PreTick;
      DoSlowTick;
      PostTick;
    end;

    procedure   TMod.DoTick; // virtual;
    begin
      // Process inputs into new outputs - audio rate
    end;

    procedure   TMod.DoSlowTick; // virtual;
    begin
      // Process inputs into new outputs - control rate
    end;

    procedure   TMod.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // virtual;
    begin
      // Collect info to be sent to module lights - this module has no lights
    end;

    procedure   TMod.GatherData( const aPrefix: string; const aCallback: TDataHandler); // virtual;
    begin
      // Collect numerical info to be sent to module viewers - this module has no viewers
    end;

    procedure   TMod.GatherInfo( const aPrefix: string; const aCallback: TInfoHandler); // virtual;
    begin
      // Collect string info to be sent to module viewers - this module has no viewers
    end;

    procedure   TMod.AcceptParam( const aPath: string; aValue: TSignal; IsFinal: Boolean);
    // Accept knob data sent from visual modules
    var
      aPrefix: string;
      aParam  : string;
      p      : Integer;
    begin
      if ( ClassType = TSynthPatch)
      then HandleParam( aPath, aValue, IsFinal)
      else begin
        ParsePrefix( aPath, aPrefix, aParam);
        if SameText( aPrefix, Name)                      // is the message for us?
        then begin
          if Pos( '/', aParam) = 0                       // When no more slashes left
          then begin                                     // we must be at the final bit ...
            p := Pos( '_', aParam);                      // The xxx_ is a conveniance in the "desiger" ... strip it here
            if p > 0
            then begin
              aParam := Copy( aParam, p + 1, Length( aParam));
              HandleParam( aParam, aValue, IsFinal);
            end;
          end;
        end;
      end;
    end;

    function    TMod.GetProfileInfoHtml: string;
    begin
      Result :=
        Format(
          '<tr><td>%s</td><td>%s</td><td align="right">%d</td><td align="right">%d</td><td align="right">%f</td></tr>',
          [
            Name,
            ClassName,
            Samples,
            TotalClocks,
            TotalClocks / Samples
          ],
          AppLocale
        );
    end;

    function    TMod.IsSameModule( const aModule: TMod): Boolean;
    begin
      Result := False;
      if Assigned( aModule)
      then Result := ClassType = aModule.ClassType;
    end;

    procedure   TMod.AcceptMidi( const aMsg: TMidiMessage); // virtual;
    begin
      // Nothing here, if is for chuldren, just building daydreams ...
    end;

    procedure   TMod.SendMidi( const aMsg: TMidiMessage);
    begin
      if Assigned( FParent)
      then FParent.SendMidi( aMsg);
    end;


{ ========
  TPatch = class
  private
    FVersion          : Integer;
    FModules          : TModuleArray;
    FConnections      : TConnections;
    FInputMappings    : TConnections;
    FOutputMappings   : TConnections;
    FListeners        : TListenerArray;
    FDecimatorCount   : Integer;
    FInputQueue       : TSignalQueue;
    FWireClocks       : Int64;
    FTotalWireClocks  : Int64;
  public
    property    Version                      : Integer     read FVersion       write FVersion;
    property    ModuleCount                  : Integer     read GetModuleCount;
    property    Module    [ anIndex: Integer]: TMod        read GetModule;
    property    ConnectionCount              : Integer     read GetConnectionCount;
    property    Overflows                    : Integer     read GetOverflows;
    property    Underflows                   : Integer     read GetUnderflows;
  private
}

    function    TSynthPatch.AssertUniqueName( const aName: string): string; // override;
    var
      i : Integer;
    begin
      // calling inherited here would tend to go recursive ...
      Result := aName;
      for i := 0 to ModuleCount - 1
      do begin
        if SameText( FModules[ i].Name, aName)
        then raise EModule.CreateFmt( 'module name "%s" is not unique.', [ aName]);
      end;
    end;

//  private

    function    TSynthPatch.CheckModuleIndex( anIndex: Integer): Boolean;
    begin
      Result := ( anIndex >= 0) and ( anIndex < ModuleCount);
    end;

    function    TSynthPatch.GetModuleCount: Integer;
    begin
      Result := Length( FModules);
    end;

    function    TSynthPatch.GetModule( anIndex: Integer): TMod;
    begin
      Result := nil;
      if CheckModuleIndex( anIndex)
      then Result := FModules[ anIndex];
    end;

    function    TSynthPatch.GetConnectionCount: Integer;
    begin
      Result := 0;
      if Assigned( FConnections)
      then Result := FConnections.Count;
    end;

    function    TSynthPatch.GetOverflows: Integer;
    begin
      if Assigned( FInputQueue)
      then Result := FInputQueue.Overflows
      else Result := 0;
    end;

    function    TSynthPatch.GetUnderflows: Integer;
    begin
      if Assigned( FInputQueue)
      then Result := FInputQueue.Underflows
      else Result := 0;
    end;

//  protected

    procedure   TSynthPatch.HandleParam( const aPath: string; aValue: TSignal; IsFinal: Boolean); // override;
    var
      i : Integer;
    begin
      for i := 0 to ModuleCount - 1
      do Module[ i].AcceptParam( aPath, aValue, IsFinal);
    end;


//  public

    constructor TSynthPatch.Create( aParent: TSynthPatch; const aName: string; anInputCount: Integer); // override;
    begin
      inherited;
      FIsSlow         := True;
      FIsFast         := True;
      FIsMidi         := True;
      FVersion        := 0;
      FInputQueue     := TSignalQueue.Create( 32768);
      FConnections    := TConnections.Create( Self);
      FInputMappings  := TConnections.Create( Self);
      FOutputMappings := TConnections.Create( Self);
    end;

    destructor  TSynthPatch.Destroy; // override;
    begin
      FreeAndNil( FConnections   );  // For a faster clear
      FreeAndNil( FOutputMappings);
      FreeAndNil( FInputMappings );
      Clear;                         // Clears the modules - that is it removes and frees them
      FreeAndNil( FInputQueue    );
      inherited;
    end;

    procedure   TSynthPatch.SetDefaults; // override;
    var
      i : Integer;
    begin
      for i := 0 to ModuleCount - 1
      do Module[ i].SetDefaults;
    end;

    procedure   TSynthPatch.AcceptText( const aValue: string); // override;
    var
      i : Integer;
    begin
      for i := 0 to Length( FListeners) - 1
      do FListeners[ i].AcceptText( aValue);
    end;

    procedure   TSynthPatch.AcceptData( const aValue: TBytes); // override;
    var
      i : Integer;
    begin
      for i := 0 to Length( FListeners) - 1
      do FListeners[ i].AcceptData( aValue);
    end;

    procedure   TSynthPatch.SaveInternals( anIndent: Integer; var S: string); // override;
    begin
      S := S + Indent( anIndent) + 'internals(' + sLineBreak;
      SaveKeyValue( anIndent, 'outputcount', OutputCount, S);
      S := S + Indent( anIndent) + ')' + sLineBreak;
    end;

    procedure   TSynthPatch.SetInternal( const aName: string; aValue: TSignal); // override;
    begin
      if SameText( aName, 'outputcount')
      then SetOutputCount( Max( 0, Ceil( aValue)))
      else inherited;
    end;

    function    TSynthPatch.GetInternal( const aName: string): TSignal; // override;
    begin
      if SameText( aName, 'outputcount')
      then Result := OutputCount
      else Result := inherited;
    end;

    procedure   TSynthPatch.SaveConnections( anIndent: Integer; var S: string);
    begin
      if Assigned( FConnections)
      then FConnections.Save( anIndent, 'connections', S);
    end;

    procedure   TSynthPatch.SaveInputMappings( anIndent: Integer; var S: string);
    begin
      if Assigned( FInputMappings)
      then FInputMappings.Save( anIndent, 'input_mappings', S);
    end;

    procedure   TSynthPatch.SaveOutputMappings( anIndent: Integer; var S: string);
    begin
      if Assigned( FOutputMappings)
      then FOutputMappings.Save( anIndent, 'output_mappings', S);
    end;

    procedure   TSynthPatch.SaveModules( anIndent: Integer; var S: string);
    var
      i : Integer;
    begin
      S := S + Indent( anIndent) + 'modules(' + sLineBreak;
      for i := 0 to ModuleCount - 1
      do Module[ i].Save( anIndent + 1, S);
      S := S + Indent( anIndent) + ')' + sLineBreak;
    end;

    procedure   TSynthPatch.SaveDetails( anIndent: Integer; var S: string); // override;
    begin
      inherited;
      S := S + Indent( anIndent) + 'version ' + IntToStr( Version) + sLineBreak;
      SaveModules       ( anIndent, S);
      SaveConnections   ( anIndent, S);
      SaveInputMappings ( anIndent, S);
      SaveOutputMappings( anIndent, S);
    end;

    procedure   TSynthPatch.Clear;   // Clear moduleas and connections
    begin
      ClearModules;
      inherited Clear;
    end;

    procedure   TSynthPatch.Reset;
    var
      i : Integer;
    begin
      for i := 0 to ModuleCount - 1
      do Module[ i].Reset;
    end;

    procedure   TSynthPatch.RemoveConnectionsFor( aModuleIndex: Integer);
    begin
      if Assigned( FConnections)
      then FConnections.RemoveConnectionsFor( aModuleIndex);
    end;

    procedure   TSynthPatch.SignalModuleRemoval( const aModule: TMod);
    var
      i : Integer;
      j : Integer;
    begin
      if aModule is TModListener
      then begin
        for i := 0 to Length( FListeners) - 1
        do begin
          if FListeners[ i] = aModule
          then begin
            for j := i + 1 to Length( FListeners) - 1
            do FListeners[ j - 1] := FListeners[ j];
            SetLength( FListeners, Length( FListeners) - 1);
            Break; // from for i ... loop
          end;
        end;
      end
    end;

    procedure   TSynthPatch.SignalModuleAddition( const aModule: TMod);
    begin
      if aModule is TModListener
      then begin
        SetLength( FListeners, Length( FListeners) + 1);
        FListeners[ Length( FListeners) - 1] := TModListener( aModule);
      end
    end;

    procedure   TSynthPatch.ClearModules;
    var
      aModule : TMod;
    begin
      while ModuleCount > 0
      do begin
        aModule :=  FModules[ ModuleCount - 1];
        SignalModuleRemoval( aModule);
        RemoveConnectionsFor( ModuleCount - 1);
        SetLength(  FModules, ModuleCount - 1);
        aModule.Free;
      end;
    end;

    function    TSynthPatch.FindModule( const aModuleName: string): Integer;
    var
      i : Integer;
    begin
      Result := UNDEF_POMOC;
      for i := 0 to ModuleCount - 1
      do begin
        if SameText( aModuleName, Module[ i].Name)
        then begin
          Result := i;
          Break;
        end;
      end;
    end;

    function    TSynthPatch.ModuleName( anIndex: Integer): string;
    begin
      if ( anIndex >= 0) and ( anIndex < ModuleCount)
      then Result := Module[ anIndex].Name
      else if anIndex = PARENT_INDEX
      then Result := 'parent'
      else Result := NO_MODULE;
    end;

    function    TSynthPatch.FindInputPin( const aModuleIndex: Integer; const aPinName: string): Integer;
    begin
      Result := UNDEF_POMOC;
      if aModuleIndex >= 0
      then Result := Module[ aModuleIndex].FindInput( aPinName);
    end;

    function    TSynthPatch.FindOutputPin( const aModuleIndex: Integer; const aPinName: string): Integer;
    begin
      Result := UNDEF_POMOC;
      if aModuleIndex >= 0
      then Result := Module[ aModuleIndex].FindOutput( aPinName);
    end;

    function    TSynthPatch.InputPinName( aModuleIndex, aPinIndex: Integer): string;
    begin
      Result := NO_INPUT;
      if ( aModuleIndex >= 0) and ( aModuleIndex < ModuleCount)
      then Result := Module[ aModuleIndex].InputName( aPinIndex)
      else if aModuleIndex = PARENT_INDEX
      then Result := Format( 'out%d', [ aPinIndex + 1], AppLocale)
    end;

    function    TSynthPatch.OutputPinName( aModuleIndex, aPinIndex: Integer): string;
    begin
      Result := NO_OUTPUT;
      if ( aModuleIndex >= 0) and ( aModuleIndex < ModuleCount)
      then Result := Module[ aModuleIndex].OutputName( aPinIndex)
      else if aModuleIndex = PARENT_INDEX
      then Result := Format( 'in%d', [ aPinIndex + 1], AppLocale)
    end;

    procedure   TSynthPatch.AddModule( const aModule: TMod); // overload;
    begin
      SetLength( FModules, ModuleCount + 1);
      FModules[ ModuleCount - 1] := aModule;
      SignalModuleAddition( aModule);
    end;

    procedure   TSynthPatch.AddConnection( aSrcModule, aSrcPin, aDstModule, aDstPin: Integer); // overload;
    begin
      if Assigned( FConnections)
      then FConnections.AddConnection( aSrcModule, aSrcPin, aDstModule, aDstPin);
    end;

    procedure   TSynthPatch.AddConnection( const aSrcModule, aSrcPin, aDstModule, aDstPin: string); // overload;
    var
      aSrcM: Integer;
      aSrcP: Integer;
      aDstM: Integer;
      aDstP: Integer;
    begin
      aSrcM := FindModule( aSrcModule);
      if aSrcM >= 0
      then begin
        aSrcP := FindOutputPin( aSrcM, aSrcPin);
        if aSrcP >= 0
        then begin
          aDstM := FindModule( aDstModule);
          if aDstM >= 0
          then begin
            aDstP := FindInputPin( aDstM, aDstPin);
            if aDstP >= 0
            then AddConnection( aSrcM, aSrcP, aDstM, aDstP)
            else raise EInputNotFound.CreateFmt( 'input "%s" not found on destination module "%s"', [ aDstPin, aDstModule]);
          end
          else raise EModuleNotFound.CreateFmt( 'destination module "%s" not found', [ aDstModule]);
        end
        else raise EOutputNotFound.CreateFmt( 'output "%s" not found on source module "%s"', [ aSrcPin, aSrcModule]);
      end
      else raise EModuleNotFound.CreateFmt( 'source module "%s" not found', [ aSrcModule]);
    end;

    procedure   TSynthPatch.AddInputMapping( aSrcPin, aDstModule, aDstPin: Integer);
    begin
      if Assigned( FInputMappings)
      then FInputMappings.AddConnection( PARENT_INDEX, aSrcPin, aDstModule, aDstPin);
    end;

    procedure   TSynthPatch.AddOutputMapping( aSrcModule, aSrcPin, aDstPin: Integer);
    begin
      if Assigned( FOutputMappings)
      then FOutputMappings.AddConnection( aSrcModule, aSrcPin, PARENT_INDEX, aDstPin);
    end;

    procedure   TSynthPatch.AddInputMapping( aSrcPin: Integer; const aDstModule, aDstPin: string); // overload;
    var
      aDstM : Integer;
      aDstP : Integer;
    begin
      while aSrcPin >= InputCount
      do begin
        FInputs.AddSignal( 0);
        FInCache := FInputs.FSignals;
      end;

      if ( aSrcPin >= 0) and ( aSrcPin < InputCount)
      then begin
        aDstM := FindModule( aDstModule);
        if aDstM >= 0
        then begin
          aDstP := FindInputPin( aDstM, aDstPin);
          if aDstP >= 0
          then AddInputMapping( aSrcPin, aDstM, aDstP)
          else raise EInputNotFound.CreateFmt( 'input "%s" not found on destination module "%s"', [ aDstPin, aDstModule]);
        end
        else raise EModuleNotFound.CreateFmt( 'destination module "%s" not found', [ aDstModule]);
      end
      else raise EInputNotFound.CreateFmt( 'internal input %d not defined for module %s', [ aSrcPin, Name])
    end;

    procedure   TSynthPatch.AddOutputMapping( const aSrcModule, aSrcPin: string; aDstPin: Integer); // overload;
    var
      aSrcM : Integer;
      aSrcP : Integer;
    begin
      aSrcM := FindModule( aSrcModule);
      if aSrcM >= 0
      then begin
        aSrcP := FindOutputPin( aSrcM, aSrcPin);
        if aSrcP >= 0
        then begin
          if ( aDstPin >= 0) and ( aDstPin < OutputCount)
          then AddOutputMapping( aSrcM, aSrcP, aDstPin)
          else raise EOutputNotFound.CreateFmt( 'internal output %d not defined for module %s', [ aDstPin, Name])
        end
        else raise EOutputNotFound.CreateFmt( 'output "%s" not found on source module "%s"', [ aSrcPin, aSrcModule]);
      end
      else raise EModuleNotFound.CreateFmt( 'source module "%s" not found', [ aSrcModule]);
    end;

    procedure   TSynthPatch.PreWires;
    begin
      {$Q-}
      FWireClocks := GetCpuClockCycleCount;
      {$Q+}
    end;

    procedure   TSynthPatch.PostWires;
    begin
      FWireClocks      := GetCpuClockCycleCount - FWireClocks;
      FTotalWireClocks := FTotalWireClocks      + FWireClocks;
    end;

    procedure   TSynthPatch.PropagateSignals;
    // Copy signals from inputs to outputs
    begin
      PreWires;
      FConnections.PropagateSignals( FModules);
      PostWires;
    end;

    procedure   TSynthPatch.ApplyInputMappings; // virtual;
    var
      i     : Integer;
      aPair : TSignalPair;
    begin
      // Copy inputs from this module over to the contained sub-module inputs
      PreWires;
      aPair := FInputQueue.GetData;
      if FInputs.Count >= 2
      then begin
        FInputs.Signal[ 0] := aPair.Left;
        FInputs.Signal[ 1] := aPair.Right;
      end;
      for i := 0 to FInputMappings.Count - 1
      do begin
        with FInputMappings.Connection[ i]
        do FModules[ DstModule].FInCache[ DstPin] := FInCache[ SrcPin];
      end;
      PostWires;
    end;

    procedure   TSynthPatch.ApplyOutputMappings; // virtual;
    var
      i : Integer;
    begin
      // Copy outputs from the contained sub modules to the outputs of this module
      PreWires;
      for i := 0 to FOutputMappings.Count - 1
      do begin
        with FOutputMappings.Connection[ i]
        do FOutCache[ DstPin] := FModules[ SrcModule].FOutCache[ SrcPin];
      end;
      PostWires;
    end;

    function    TSynthPatch.InternalName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 :  Result := 'outputcount';
        else Result := inherited;
      end;
    end;

    procedure   TSynthPatch.DoTick; // Override;
    var
      i  : Integer;
      p  : Integer;
      MC : Integer;
    begin
      MC := ModuleCount;
      FDecimatorCount := ( FDecimatorCount + 1) mod SLOW_DECIMATION;
      ApplyInputMappings;                     // Get patch external inputs
      p := 0;
      for i := 0 to MC - 1                    // Tick the modules in the patch
      do begin
        with FModules[ i]
        do begin
          if FIsFast
          then Tick;
          if FIsSlow
          then begin
            if p mod SLOW_DECIMATION = FDecimatorCount
            then SlowTick;
            Inc( p);
          end
        end;
      end;
      PropagateSignals;                       // Transfer module outputs to module inputs
      ApplyOutputMappings;                    // Set the patch external outputs
    end;

    procedure   TSynthPatch.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    var
      i  : Integer;
    begin
      if aPrefix = ''
      then begin
        for i := 0 to ModuleCount - 1
        do Module[ i].GatherLights( Name, aCallback);
      end
      else begin
        for i := 0 to ModuleCount - 1
        do Module[ i].GatherLights( aPrefix + '/' + Name, aCallback);
      end;
    end;

    procedure   TSynthPatch.GatherData( const aPrefix: string; const aCallback: TDataHandler); // override;
    var
      i  : Integer;
    begin
      if aPrefix = ''
      then begin
        for i := 0 to ModuleCount - 1
        do Module[ i].GatherData( Name, aCallback);
      end
      else begin
        for i := 0 to ModuleCount - 1
        do Module[ i].GatherData( aPrefix + '/' + Name, aCallback);
      end;
    end;

    procedure   TSynthPatch.GatherInfo( const aPrefix: string; const aCallback: TInfoHandler); // override;
    var
      i  : Integer;
    begin
      if aPrefix = ''
      then begin
        for i := 0 to ModuleCount - 1
        do Module[ i].GatherInfo( Name, aCallback);
      end
      else begin
        for i := 0 to ModuleCount - 1
        do Module[ i].GatherInfo( aPrefix + '/' + Name, aCallback);
      end;
    end;

    procedure   TSynthPatch.AcceptSamplePair( aLeft, aRight: TSignal);
    begin
      FInputQueue.Append( aLeft, aRight);
    end;

    function    TSynthPatch.GetProfileHtml: TStringList;
    var
      i          : Integer;
      SumClocks  : Int64;
      SumSamples : Int64;
    begin
      SumClocks  := 0;
      SumSamples := 0;
      Result := TStringList.Create;
      Result.Add( '<!doctype html>');
      Result.Add( '<html><body><table><tbody>');
      Result.Add( '<tr><td>name</td><td>type</td><td>samples</td><td>clocks</td><td>clocks / sample</td></tr>');
      for i := 0 to ModuleCount - 1
      do begin
        Result.Add( FModules[ i].GetProfileInfoHtml);
        Inc( SumClocks , FModules[ i].TotalClocks);
        Inc( SumSamples, FModules[ i].Samples    );
      end;
      Result.Add( GetProfileInfoHtml);
      Result.Add(
        Format(
          '<tr><td>%s</td><td>%s</td><td align="right">%d</td><td align="right">%d</td><td align="right">%f</td></tr>',
          [
            'total',
            'average module',
            SumSamples,
            SumClocks,
            SumClocks / SumSamples
          ],
          AppLocale
        )
      );
      Result.Add(
        Format(
          '<tr><td>%s</td><td>%s</td><td align="right">%d</td><td align="right">%d</td><td align="right">%f</td></tr>',
          [
            'wires',
            '&nbsp;',
            Samples,
            FTotalWireClocks,
            FTotalWireClocks / Samples
          ],
          AppLocale
        )
      );
      Result.Add(
        Format(
          '<tr><td>%s</td><td>%s</td><td align="right">%d</td><td align="right">%f</td><td align="right">%f</td></tr>',
          [
            'wire',
            'average',
            Samples,
            FTotalWireClocks / ConnectionCount,
            FTotalWireClocks / ConnectionCount / Samples
          ],
          AppLocale
        )
      );
      Result.Add(
        Format(
          '<tr><td>%s</td><td>%s</td><td align="right">%d</td><td align="right">%d</td><td align="right">%f</td></tr>',
          [
            'overhead',
            'lights etc.',
            Samples,
            TotalClocks - SumClocks - FTotalWireClocks,
            ( TotalClocks - SumClocks - FTotalWireClocks) / Samples
          ],
          AppLocale
        )
      );
      Result.Add( '<tr><td>name</td><td>type</td><td>samples</td><td>clocks</td><td>clocks / sample</td></tr>');
      Result.Add( '</tbody></table></body></html>');
    end;

    function    TSynthPatch.IsSamePatch( const aPatch: TSynthPatch): Boolean;
    begin
      Result := False;
      if Assigned( aPatch)
      then begin
        Result :=
          SameModules       ( aPatch) and
          SameConnections   ( aPatch) and
          SameInputMappings ( aPatch) and
          SameOutputMappings( aPatch);
      end;
    end;

    function    TSynthPatch.SameModules( const aPatch: TSynthPatch): Boolean;
    var
      i : Integer;
    begin
      Result := False;
      if Assigned( aPatch)
      then begin
        if ModuleCount = aPatch.ModuleCount
        then begin
          Result := True;
          for i := 0 to ModuleCount - 1
          do begin
            if not Module[ i].IsSameModule( aPatch.Module[ i])
            then begin
              Result := False;
              Break;
            end;
          end;
        end;
      end;
    end;

    function    TSynthPatch.SameConnections( const aPatch: TSynthPatch): Boolean;
    begin
      Result := False;
      if Assigned( aPatch)
      then Result := FConnections.IsSameConnections( aPatch.FConnections);
    end;

    function    TSynthPatch.SameInputMappings( const aPatch: TSynthPatch): Boolean;
    begin
      Result := False;
      if Assigned( aPatch)
      then Result := FInputMappings.IsSameConnections( aPatch.FInputMappings);
    end;

    function    TSynthPatch.SameOutputMappings( const aPatch: TSynthPatch): Boolean;
    begin
      Result := False;
      if Assigned( aPatch)
      then Result := FOutputMappings.IsSameConnections( aPatch.FOutputMappings);
    end;

    procedure   TSynthPatch.AcceptMidi( const aMsg: TMidiMessage); // overide;
    var
      i : Integer;
    begin
      if FIsMidi
      then begin
        for i := 0 to ModuleCount - 1
        do begin
          if Module[ i].FIsMidi
          then Module[ i].AcceptMidi( aMsg);
        end;
      end;
    end;

    procedure   TSynthPatch.SendMidi( const aMsg: TMidiMessage); // override;
    begin
      if Assigned( FOnSendMidi)
      then FOnSendMidi( Self, aMsg);
    end;


initialization

  InitializeUnit;

finalization

  FinalizeUnit;

end.
